@@ -117,50 +117,6 @@ static inline const char *network_operator_status_to_string(int status)
 	return "unknown";
 }
 
-static inline const char *registration_status_to_string(int status)
-{
-	switch (status) {
-	case NETWORK_REGISTRATION_STATUS_NOT_REGISTERED:
-		return "unregistered";
-	case NETWORK_REGISTRATION_STATUS_REGISTERED:
-		return "registered";
-	case NETWORK_REGISTRATION_STATUS_SEARCHING:
-		return "searching";
-	case NETWORK_REGISTRATION_STATUS_DENIED:
-		return "denied";
-	case NETWORK_REGISTRATION_STATUS_UNKNOWN:
-		return "unknown";
-	case NETWORK_REGISTRATION_STATUS_ROAMING:
-		return "roaming";
-	}
-
-	return "";
-}
-
-static inline const char *registration_tech_to_string(int tech)
-{
-	switch (tech) {
-	case ACCESS_TECHNOLOGY_GSM:
-		return "GSM";
-	case ACCESS_TECHNOLOGY_GSM_COMPACT:
-		return "GSMCompact";
-	case ACCESS_TECHNOLOGY_UTRAN:
-		return "UTRAN";
-	case ACCESS_TECHNOLOGY_GSM_EGPRS:
-		return "GSM+EGPRS";
-	case ACCESS_TECHNOLOGY_UTRAN_HSDPA:
-		return "UTRAN+HSDPA";
-	case ACCESS_TECHNOLOGY_UTRAN_HSUPA:
-		return "UTRAN+HSUPA";
-	case ACCESS_TECHNOLOGY_UTRAN_HSDPA_HSUPA:
-		return "UTRAN+HSDPA+HSUPA";
-	case ACCESS_TECHNOLOGY_EUTRAN:
-		return "EUTRAN";
-	default:
-		return "";
-	}
-}
-
 static void register_callback(const struct ofono_error *error, void *data)
 {
 	struct ofono_netreg *netreg = data;
