@@ -57,6 +57,26 @@ gboolean __ofono_dbus_valid_object_path(const char *path);
 
 #include <ofono/types.h>
 
+struct ofono_watchlist_item {
+	unsigned int id;
+	void *notify;
+	void *notify_data;
+	ofono_destroy_func destroy;
+};
+
+struct ofono_watchlist {
+	int next_id;
+	GSList *items;
+	ofono_destroy_func destroy;
+};
+
+struct ofono_watchlist *__ofono_watchlist_new(ofono_destroy_func destroy);
+unsigned int __ofono_watchlist_add_item(struct ofono_watchlist *watchlist,
+					struct ofono_watchlist_item *item);
+gboolean __ofono_watchlist_remove_item(struct ofono_watchlist *watchlist,
+					unsigned int id);
+void __ofono_watchlist_free(struct ofono_watchlist *watchlist);
+
 #include <ofono/plugin.h>
 
 int __ofono_plugin_init(const char *pattern, const char *exclude);
@@ -84,6 +104,8 @@ enum ofono_atom_type {
 	OFONO_ATOM_TYPE_HISTORY = 11,
 	OFONO_ATOM_TYPE_SSN = 12,
 	OFONO_ATOM_TYPE_MESSAGE_WAITING = 13,
+	OFONO_ATOM_TYPE_CBS = 14,
+	OFONO_ATOM_TYPES_CALL_VOLUME = 15,
 };
 
 enum ofono_atom_watch_condition {
@@ -119,11 +141,12 @@ void __ofono_atom_unregister(struct ofono_atom *atom);
 
 gboolean __ofono_atom_get_registered(struct ofono_atom *atom);
 
-int __ofono_modem_add_atom_watch(struct ofono_modem *modem,
+unsigned int __ofono_modem_add_atom_watch(struct ofono_modem *modem,
 					enum ofono_atom_type type,
 					ofono_atom_watch_func notify,
 					void *data, ofono_destroy_func destroy);
-gboolean __ofono_modem_remove_atom_watch(struct ofono_modem *modem, int id);
+gboolean __ofono_modem_remove_atom_watch(struct ofono_modem *modem,
+						unsigned int id);
 
 void __ofono_atom_free(struct ofono_atom *atom);
 
@@ -131,6 +154,7 @@ void __ofono_atom_free(struct ofono_atom *atom);
 #include <ofono/call-forwarding.h>
 #include <ofono/call-meter.h>
 #include <ofono/call-settings.h>
+#include <ofono/cbs.h>
 #include <ofono/devinfo.h>
 #include <ofono/phonebook.h>
 #include <ofono/sms.h>
@@ -178,6 +202,20 @@ void __ofono_ussd_passwd_unregister(struct ofono_ussd *ussd, const char *sc);
 
 #include <ofono/netreg.h>
 
+typedef void (*ofono_netreg_status_notify_cb_t)(int status, int lac, int ci,
+			int tech, const struct ofono_network_operator *op,
+			void *data);
+
+unsigned int __ofono_netreg_add_status_watch(struct ofono_netreg *netreg,
+				ofono_netreg_status_notify_cb_t cb,
+				void *data, ofono_destroy_func destroy);
+
+gboolean __ofono_netreg_remove_status_watch(struct ofono_netreg *netreg,
+						unsigned int id);
+
+void __ofono_netreg_set_base_station_name(struct ofono_netreg *netreg,
+						const char *name);
+
 #include <ofono/history.h>
 
 void __ofono_history_probe_drivers(struct ofono_modem *modem);
