@@ -25,19 +25,15 @@
 
 #include <string.h>
 #include <stdio.h>
+#include <errno.h>
 
-#include <dbus/dbus.h>
 #include <glib.h>
 #include <gdbus.h>
 
 #include "ofono.h"
 
-#include "dbus-gsm.h"
-#include "modem.h"
-#include "driver.h"
 #include "common.h"
 #include "util.h"
-#include "sim.h"
 #include "smsutil.h"
 
 #define uninitialized_var(x) x = x
@@ -48,8 +44,9 @@
 
 static gboolean tx_next(gpointer user_data);
 
-struct sms_manager_data {
-	struct ofono_sms_ops *ops;
+static GSList *g_drivers = NULL;
+
+struct ofono_sms {
 	int flags;
 	DBusMessage *pending;
 	struct ofono_phone_number sca;
@@ -57,6 +54,13 @@ struct sms_manager_data {
 	guint ref;
 	GQueue *txq;
 	time_t last_mms;
+	gint tx_source;
+	struct ofono_message_waiting *mw;
+	unsigned int mw_watch;
+	struct ofono_sim *sim;
+	const struct ofono_sms_driver *driver;
+	void *driver_data;
+	struct ofono_atom *atom;
 };
 
 struct pending_pdu {
@@ -65,45 +69,11 @@ struct pending_pdu {
 	int pdu_len;
 };
 
-static struct sms_manager_data *sms_manager_create()
-{
-	struct sms_manager_data *sms;
-
-	sms = g_new0(struct sms_manager_data, 1);
-
-	sms->sca.type = 129;
-	sms->ref = 1;
-
-	sms->assembly = sms_assembly_new();
-	sms->txq = g_queue_new();
-
-	return sms;
-}
-
-static void sms_manager_destroy(gpointer userdata)
-{
-	struct ofono_modem *modem = userdata;
-	struct sms_manager_data *data = modem->sms_manager;
-
-	if (data->assembly) {
-		sms_assembly_free(data->assembly);
-		data->assembly = NULL;
-	}
-
-	if (data->txq) {
-		g_queue_foreach(data->txq, (GFunc)g_free, NULL);
-		g_queue_free(data->txq);
-		data->txq = NULL;
-	}
-
-	g_free(data);
-}
-
-static void set_sca(struct ofono_modem *modem,
+static void set_sca(struct ofono_sms *sms,
 			const struct ofono_phone_number *sca)
 {
-	struct sms_manager_data *sms = modem->sms_manager;
-	DBusConnection *conn = dbus_gsm_connection();
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(sms->atom);
 	const char *value;
 
 	if (sms->sca.type == sca->type &&
@@ -116,16 +86,15 @@ static void set_sca(struct ofono_modem *modem,
 
 	value = phone_number_to_string(&sms->sca);
 
-	dbus_gsm_signal_property_changed(conn, modem->path,
+	ofono_dbus_signal_property_changed(conn, path,
 						SMS_MANAGER_INTERFACE,
 						"ServiceCenterAddress",
 						DBUS_TYPE_STRING, &value);
 }
 
-static DBusMessage *generate_get_properties_reply(struct ofono_modem *modem,
+static DBusMessage *generate_get_properties_reply(struct ofono_sms *sms,
 							DBusMessage *msg)
 {
-	struct sms_manager_data *sms = modem->sms_manager;
 	DBusMessage *reply;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
@@ -139,12 +108,12 @@ static DBusMessage *generate_get_properties_reply(struct ofono_modem *modem,
 	dbus_message_iter_init_append(reply, &iter);
 
 	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
-						PROPERTIES_ARRAY_SIGNATURE,
+					OFONO_PROPERTIES_ARRAY_SIGNATURE,
 						&dict);
 
 	sca = phone_number_to_string(&sms->sca);
 
-	dbus_gsm_dict_append(&dict, "ServiceCenterAddress", DBUS_TYPE_STRING,
+	ofono_dbus_dict_append(&dict, "ServiceCenterAddress", DBUS_TYPE_STRING,
 				&sca);
 
 	dbus_message_iter_close_container(&iter, &dict);
@@ -155,42 +124,40 @@ static DBusMessage *generate_get_properties_reply(struct ofono_modem *modem,
 static void sms_sca_query_cb(const struct ofono_error *error,
 				const struct ofono_phone_number *sca, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct sms_manager_data *sms = modem->sms_manager;
+	struct ofono_sms *sms = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
 		goto out;
 
-	set_sca(modem, sca);
+	set_sca(sms, sca);
 
 	sms->flags |= SMS_MANAGER_FLAG_CACHED;
 
 out:
 	if (sms->pending) {
-		DBusMessage *reply = generate_get_properties_reply(modem,
+		DBusMessage *reply = generate_get_properties_reply(sms,
 								sms->pending);
-		dbus_gsm_pending_reply(&sms->pending, reply);
+		__ofono_dbus_pending_reply(&sms->pending, reply);
 	}
 }
 
 static DBusMessage *sms_get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct sms_manager_data *sms = modem->sms_manager;
+	struct ofono_sms *sms = data;
 
 	if (sms->pending)
-		return dbus_gsm_busy(msg);
+		return __ofono_error_busy(msg);
 
-	if (!sms->ops->sca_query)
-		return dbus_gsm_not_implemented(msg);
+	if (!sms->driver->sca_query)
+		return __ofono_error_not_implemented(msg);
 
 	if (sms->flags & SMS_MANAGER_FLAG_CACHED)
-		return generate_get_properties_reply(modem, msg);
+		return generate_get_properties_reply(sms, msg);
 
 	sms->pending = dbus_message_ref(msg);
 
-	sms->ops->sca_query(modem, sms_sca_query_cb, modem);
+	sms->driver->sca_query(sms, sms_sca_query_cb, sms);
 
 	return NULL;
 }
@@ -199,62 +166,59 @@ static void sca_set_query_callback(const struct ofono_error *error,
 					const struct ofono_phone_number *sca,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct sms_manager_data *sms = modem->sms_manager;
+	struct ofono_sms *sms = data;
 	DBusMessage *reply;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		ofono_error("Set SCA succeeded, but query failed");
 		sms->flags &= ~SMS_MANAGER_FLAG_CACHED;
-		reply = dbus_gsm_failed(sms->pending);
-		dbus_gsm_pending_reply(&sms->pending, reply);
+		reply = __ofono_error_failed(sms->pending);
+		__ofono_dbus_pending_reply(&sms->pending, reply);
 		return;
 	}
 
-	set_sca(modem, sca);
+	set_sca(sms, sca);
 
 	reply = dbus_message_new_method_return(sms->pending);
-	dbus_gsm_pending_reply(&sms->pending, reply);
+	__ofono_dbus_pending_reply(&sms->pending, reply);
 }
 
 static void sca_set_callback(const struct ofono_error *error, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct sms_manager_data *sms = modem->sms_manager;
+	struct ofono_sms *sms = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		ofono_debug("Setting SCA failed");
-		dbus_gsm_pending_reply(&sms->pending,
-					dbus_gsm_failed(sms->pending));
+		__ofono_dbus_pending_reply(&sms->pending,
+					__ofono_error_failed(sms->pending));
 		return;
 	}
 
-	sms->ops->sca_query(modem, sca_set_query_callback, modem);
+	sms->driver->sca_query(sms, sca_set_query_callback, sms);
 }
 
 static DBusMessage *sms_set_property(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct sms_manager_data *sms = modem->sms_manager;
+	struct ofono_sms *sms = data;
 	DBusMessageIter iter;
 	DBusMessageIter var;
 	const char *property;
 
 	if (sms->pending)
-		return dbus_gsm_busy(msg);
+		return __ofono_error_busy(msg);
 
 	if (!dbus_message_iter_init(msg, &iter))
-		return dbus_gsm_invalid_args(msg);
+		return __ofono_error_invalid_args(msg);
 
 	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
-		return dbus_gsm_invalid_args(msg);
+		return __ofono_error_invalid_args(msg);
 
 	dbus_message_iter_get_basic(&iter, &property);
 	dbus_message_iter_next(&iter);
 
 	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
-		return dbus_gsm_invalid_args(msg);
+		return __ofono_error_invalid_args(msg);
 
 	dbus_message_iter_recurse(&iter, &var);
 
@@ -263,38 +227,37 @@ static DBusMessage *sms_set_property(DBusConnection *conn, DBusMessage *msg,
 		struct ofono_phone_number sca;
 
 		if (dbus_message_iter_get_arg_type(&var) != DBUS_TYPE_STRING)
-			return dbus_gsm_invalid_args(msg);
+			return __ofono_error_invalid_args(msg);
 
 		dbus_message_iter_get_basic(&var, &value);
 
 		if (strlen(value) == 0 || !valid_phone_number_format(value))
-			return dbus_gsm_invalid_format(msg);
+			return __ofono_error_invalid_format(msg);
 
-		if (!sms->ops->sca_set)
-			return dbus_gsm_not_implemented(msg);
+		if (!sms->driver->sca_set)
+			return __ofono_error_not_implemented(msg);
 
 		string_to_phone_number(value, &sca);
 
 		sms->pending = dbus_message_ref(msg);
 
-		sms->ops->sca_set(modem, &sca, sca_set_callback, modem);
+		sms->driver->sca_set(sms, &sca, sca_set_callback, sms);
 		return NULL;
 	}
 
-	return dbus_gsm_invalid_args(msg);
+	return __ofono_error_invalid_args(msg);
 }
 
 static void tx_finished(const struct ofono_error *error, int mr, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct sms_manager_data *sms = modem->sms_manager;
+	struct ofono_sms *sms = data;
 	struct pending_pdu *pdu;
 
 	ofono_debug("tx_finished");
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		ofono_debug("Sending failed, retrying in 5 seconds...");
-		g_timeout_add_seconds(5, tx_next, data);
+		sms->tx_source = g_timeout_add_seconds(5, tx_next, sms);
 		return;
 	}
 
@@ -305,16 +268,15 @@ static void tx_finished(const struct ofono_error *error, int mr, void *data)
 
 	if (g_queue_peek_head(sms->txq)) {
 		ofono_debug("Scheduling next");
-		g_timeout_add(0, tx_next, sms);
+		sms->tx_source = g_timeout_add(0, tx_next, sms);
 	}
 }
 
 static gboolean tx_next(gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct sms_manager_data *sms = modem->sms_manager;
+	struct ofono_sms *sms = user_data;
 	time_t ts;
-	gboolean send_mms = FALSE;
+	int send_mms = 0;
 	struct pending_pdu *pdu = g_queue_peek_head(sms->txq);
 	struct ofono_error error;
 
@@ -322,6 +284,8 @@ static gboolean tx_next(gpointer user_data)
 
 	ofono_debug("tx_next: %p", pdu);
 
+	sms->tx_source = 0;
+
 	if (!pdu)
 		return FALSE;
 
@@ -329,10 +293,10 @@ static gboolean tx_next(gpointer user_data)
 
 	if ((g_queue_get_length(sms->txq) > 1) &&
 			((ts - sms->last_mms) > 60))
-		send_mms = TRUE;
+		send_mms = 1;
 
-	sms->ops->submit(modem, pdu->pdu, pdu->pdu_len, pdu->tpdu_len, send_mms,
-				tx_finished, modem);
+	sms->driver->submit(sms, pdu->pdu, pdu->pdu_len, pdu->tpdu_len, send_mms,
+				tx_finished, sms);
 
 	return FALSE;
 }
@@ -355,9 +319,8 @@ static void set_ref_and_to(GSList *msg_list, guint16 ref, int offset,
 	}
 }
 
-static void append_tx_queue(struct ofono_modem *modem, GSList *msg_list)
+static void append_tx_queue(struct ofono_sms *sms, GSList *msg_list)
 {
-	struct sms_manager_data *sms = modem->sms_manager;
 	struct sms *s;
 	GSList *l;
 	struct pending_pdu *pdu;
@@ -380,14 +343,13 @@ static void append_tx_queue(struct ofono_modem *modem, GSList *msg_list)
 	}
 
 	if (start)
-		g_timeout_add(0, tx_next, modem);
+		sms->tx_source = g_timeout_add(0, tx_next, sms);
 }
 
 static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct sms_manager_data *sms = modem->sms_manager;
+	struct ofono_sms *sms = data;
 	char **tos;
 	int num_to;
 	char *text;
@@ -398,11 +360,11 @@ static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
 					&tos, &num_to, DBUS_TYPE_STRING, &text,
 					DBUS_TYPE_INVALID))
-		return dbus_gsm_invalid_args(msg);
+		return __ofono_error_invalid_args(msg);
 
 	if (num_to == 0) {
 		dbus_free_string_array(tos);
-		return dbus_gsm_invalid_format(msg);
+		return __ofono_error_invalid_format(msg);
 	}
 
 	ofono_debug("Got %d recipients", num_to);
@@ -412,20 +374,20 @@ static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 			continue;
 
 		dbus_free_string_array(tos);
-		return dbus_gsm_invalid_format(msg);
+		return __ofono_error_invalid_format(msg);
 	}
 
 	msg_list = sms_text_prepare(text, 0, TRUE, &ref_offset);
 
 	if (!msg_list) {
 		dbus_free_string_array(tos);
-		return dbus_gsm_invalid_format(msg);
+		return __ofono_error_invalid_format(msg);
 	}
 
 	for (i = 0; i < num_to; i++) {
 		ofono_debug("ref: %d, offset: %d", sms->ref, ref_offset);
 		set_ref_and_to(msg_list, sms->ref, ref_offset, tos[i]);
-		append_tx_queue(modem, msg_list);
+		append_tx_queue(sms, msg_list);
 
 		if (sms->ref == 65536)
 			sms->ref = 1;
@@ -456,7 +418,7 @@ static GDBusSignalTable sms_manager_signals[] = {
 	{ }
 };
 
-static void dispatch_app_datagram(struct ofono_modem *modem, int dst, int src,
+static void dispatch_app_datagram(struct ofono_sms *sms, int dst, int src,
 					unsigned char *buf, long len)
 {
 	ofono_debug("Got app datagram for dst port: %d, src port: %d",
@@ -464,13 +426,14 @@ static void dispatch_app_datagram(struct ofono_modem *modem, int dst, int src,
 	ofono_debug("Contents-Len: %ld", len);
 }
 
-static void dispatch_text_message(struct ofono_modem *modem,
+static void dispatch_text_message(struct ofono_sms *sms,
 					const char *message,
 					enum sms_class cls,
 					const struct sms_address *addr,
 					const struct sms_scts *scts)
 {
-	DBusConnection *conn = dbus_gsm_connection();
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(sms->atom);
 	DBusMessage *signal;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
@@ -488,7 +451,7 @@ static void dispatch_text_message(struct ofono_modem *modem,
 	else
 		signal_name = "IncomingMessage";
 
-	signal = dbus_message_new_signal(modem->path, SMS_MANAGER_INTERFACE,
+	signal = dbus_message_new_signal(path, SMS_MANAGER_INTERFACE,
 						signal_name);
 
 	if (!signal)
@@ -499,31 +462,31 @@ static void dispatch_text_message(struct ofono_modem *modem,
 	dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &message);
 
 	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
-						PROPERTIES_ARRAY_SIGNATURE,
+					OFONO_PROPERTIES_ARRAY_SIGNATURE,
 						&dict);
 
 	ts = sms_scts_to_time(scts, &remote);
 
 	strftime(buf, 127, "%a, %d %b %Y %H:%M:%S %z", localtime(&ts));
 	buf[127] = '\0';
-	dbus_gsm_dict_append(&dict, "LocalSentTime", DBUS_TYPE_STRING, &str);
+	ofono_dbus_dict_append(&dict, "LocalSentTime", DBUS_TYPE_STRING, &str);
 
 	strftime(buf, 127, "%a, %d %b %Y %H:%M:%S %z", &remote);
 	buf[127] = '\0';
-	dbus_gsm_dict_append(&dict, "SentTime", DBUS_TYPE_STRING, &str);
+	ofono_dbus_dict_append(&dict, "SentTime", DBUS_TYPE_STRING, &str);
 
 	str = sms_address_to_string(addr);
-	dbus_gsm_dict_append(&dict, "Sender", DBUS_TYPE_STRING, &str);
+	ofono_dbus_dict_append(&dict, "Sender", DBUS_TYPE_STRING, &str);
 
 	dbus_message_iter_close_container(&iter, &dict);
 
 	g_dbus_send_message(conn, signal);
 }
 
-static void sms_dispatch(struct ofono_modem *modem, GSList *sms_list)
+static void sms_dispatch(struct ofono_sms *sms, GSList *sms_list)
 {
 	GSList *l;
-	const struct sms *sms;
+	const struct sms *s;
 	enum sms_charset uninitialized_var(old_charset);
 	enum sms_class cls;
 	int srcport = -1;
@@ -553,8 +516,8 @@ static void sms_dispatch(struct ofono_modem *modem, GSList *sms_list)
 		int csrc = -1;
 		gboolean is_8bit;
 
-		sms = l->data;
-		dcs = sms->deliver.dcs;
+		s = l->data;
+		dcs = s->deliver.dcs;
 
 		if (sms_mwi_dcs_decode(dcs, NULL, &charset, NULL, NULL))
 			cls = SMS_CLASS_UNSPECIFIED;
@@ -576,7 +539,7 @@ static void sms_dispatch(struct ofono_modem *modem, GSList *sms_list)
 			return;
 		}
 
-		if (sms_extract_app_port(sms, &cdst, &csrc, &is_8bit) &&
+		if (sms_extract_app_port(s, &cdst, &csrc, &is_8bit) &&
 				(l == sms_list)) {
 			srcport = is_8bit ? csrc : (csrc << 8);
 			dstport = is_8bit ? cdst : (cdst << 8);
@@ -606,7 +569,7 @@ static void sms_dispatch(struct ofono_modem *modem, GSList *sms_list)
 		if (!buf)
 			return;
 
-		dispatch_app_datagram(modem, dstport, srcport, buf, len);
+		dispatch_app_datagram(sms, dstport, srcport, buf, len);
 
 		g_free(buf);
 	} else {
@@ -615,18 +578,16 @@ static void sms_dispatch(struct ofono_modem *modem, GSList *sms_list)
 		if (!message)
 			return;
 
-		sms = sms_list->data;
+		s = sms_list->data;
 
-		dispatch_text_message(modem, message, cls, &sms->deliver.oaddr,
-					&sms->deliver.scts);
+		dispatch_text_message(sms, message, cls, &s->deliver.oaddr,
+					&s->deliver.scts);
 		g_free(message);
 	}
 }
 
-static void handle_deliver(struct ofono_modem *modem,
-				const struct sms *incoming)
+static void handle_deliver(struct ofono_sms *sms, const struct sms *incoming)
 {
-	struct sms_manager_data *sms = modem->sms_manager;
 	GSList *l;
 	guint16 ref;
 	guint8 max;
@@ -646,7 +607,7 @@ static void handle_deliver(struct ofono_modem *modem,
 		if (!sms_list)
 			return;
 
-		sms_dispatch(modem, sms_list);
+		sms_dispatch(sms, sms_list);
 		g_slist_foreach(sms_list, (GFunc)g_free, NULL);
 		g_slist_free(sms_list);
 
@@ -654,61 +615,67 @@ static void handle_deliver(struct ofono_modem *modem,
 	}
 
 	l = g_slist_append(NULL, (void *)incoming);
-	sms_dispatch(modem, l);
+	sms_dispatch(sms, l);
 	g_slist_free(l);
 }
 
-static void handle_mwi(struct ofono_modem *modem, struct sms *mwi)
+static inline gboolean handle_mwi(struct ofono_sms *sms, struct sms *s)
 {
-	ofono_error("MWI information not yet handled");
-}
-
-void ofono_sms_deliver_notify(struct ofono_modem *modem, unsigned char *pdu,
-				int len, int tpdu_len)
-{
-	struct sms sms;
-	enum sms_class cls;
 	gboolean discard;
 
-	if (!sms_decode(pdu, len, FALSE, tpdu_len, &sms)) {
+	if (sms->mw == NULL)
+		return FALSE;
+
+	__ofono_message_waiting_mwi(sms->mw, s, &discard);
+
+	return discard;
+}
+
+void ofono_sms_deliver_notify(struct ofono_sms *sms, unsigned char *pdu,
+				int len, int tpdu_len)
+{
+	struct sms s;
+	enum sms_class cls;
+
+	if (!sms_decode(pdu, len, FALSE, tpdu_len, &s)) {
 		ofono_error("Unable to decode PDU");
 		return;
 	}
 
-	if (sms.type != SMS_TYPE_DELIVER) {
+	if (s.type != SMS_TYPE_DELIVER) {
 		ofono_error("Expecting a DELIVER pdu");
 		return;
 	}
 
-	if (sms.deliver.pid == SMS_PID_TYPE_SM_TYPE_0) {
+	if (s.deliver.pid == SMS_PID_TYPE_SM_TYPE_0) {
 		ofono_debug("Explicitly ignoring type 0 SMS");
 		return;
 	}
 
 	/* This is an older style MWI notification, process MWI
 	 * headers and handle it like any other message */
-	if (sms.deliver.pid == SMS_PID_TYPE_RETURN_CALL) {
-		handle_mwi(modem, &sms);
-		goto out;
-	}
-
-	/* The DCS indicates this is an MWI notification, process it
-	 * and then handle the User-Data as any other message */
-	if (sms_mwi_dcs_decode(sms.deliver.dcs, NULL, NULL, NULL, &discard)) {
-		handle_mwi(modem, &sms);
-
-		if (discard)
+	if (s.deliver.pid == SMS_PID_TYPE_RETURN_CALL) {
+		if (handle_mwi(sms, &s))
 			return;
 
 		goto out;
 	}
 
-	if (!sms_dcs_decode(sms.deliver.dcs, &cls, NULL, NULL, NULL)) {
+	/* The DCS indicates this is an MWI notification, process it
+	 * and then handle the User-Data as any other message */
+	if (sms_mwi_dcs_decode(s.deliver.dcs, NULL, NULL, NULL, NULL)) {
+		if (handle_mwi(sms, &s))
+			return;
+
+		goto out;
+	}
+
+	if (!sms_dcs_decode(s.deliver.dcs, &cls, NULL, NULL, NULL)) {
 		ofono_error("Unknown / Reserved DCS.  Ignoring");
 		return;
 	}
 
-	switch (sms.deliver.pid) {
+	switch (s.deliver.pid) {
 	case SMS_PID_TYPE_ME_DOWNLOAD:
 		if (cls == SMS_CLASS_1) {
 			ofono_error("ME Download message ignored");
@@ -717,7 +684,7 @@ void ofono_sms_deliver_notify(struct ofono_modem *modem, unsigned char *pdu,
 
 		break;
 	case SMS_PID_TYPE_ME_DEPERSONALIZATION:
-		if (sms.deliver.dcs == 0x11) {
+		if (s.deliver.dcs == 0x11) {
 			ofono_error("ME Depersonalization message ignored");
 			return;
 		}
@@ -742,11 +709,11 @@ void ofono_sms_deliver_notify(struct ofono_modem *modem, unsigned char *pdu,
 	 * WCMP headers or headers that can't possibly be in a normal
 	 * message.  If we find messages like that, we ignore them.
 	 */
-	if (sms.deliver.udhi) {
+	if (s.deliver.udhi) {
 		struct sms_udh_iter iter;
 		enum sms_iei iei;
 
-		if (!sms_udh_iter_init(&sms, &iter))
+		if (!sms_udh_iter_init(&s, &iter))
 			goto out;
 
 		while ((iei = sms_udh_iter_get_ie_type(&iter)) !=
@@ -760,7 +727,12 @@ void ofono_sms_deliver_notify(struct ofono_modem *modem, unsigned char *pdu,
 			switch (iei) {
 			case SMS_IEI_SPECIAL_MESSAGE_INDICATION:
 			case SMS_IEI_ENHANCED_VOICE_MAIL_INFORMATION:
-				handle_mwi(modem, &sms);
+				/* TODO: ignore if not in the very first
+				 * segment of a concatenated SM so as not
+				 * to repeat the indication.  */
+				if (handle_mwi(sms, &s))
+					return;
+
 				goto out;
 			case SMS_IEI_WCMP:
 				ofono_error("No support for WCMP, ignoring");
@@ -772,59 +744,185 @@ void ofono_sms_deliver_notify(struct ofono_modem *modem, unsigned char *pdu,
 	}
 
 out:
-	handle_deliver(modem, &sms);
+	handle_deliver(sms, &s);
 }
 
-void ofono_sms_status_notify(struct ofono_modem *modem, unsigned char *pdu,
+void ofono_sms_status_notify(struct ofono_sms *sms, unsigned char *pdu,
 				int len, int tpdu_len)
 {
 	ofono_error("SMS Status-Report not yet handled");
 }
 
-int ofono_sms_manager_register(struct ofono_modem *modem,
-					struct ofono_sms_ops *ops)
+int ofono_sms_driver_register(const struct ofono_sms_driver *d)
 {
-	DBusConnection *conn = dbus_gsm_connection();
+	DBG("driver: %p, name: %s", d, d->name);
 
-	if (modem == NULL)
-		return -1;
+	if (d->probe == NULL)
+		return -EINVAL;
 
-	if (ops == NULL)
-		return -1;
-
-	modem->sms_manager = sms_manager_create();
-
-	if (!modem->sms_manager)
-		return -1;
-
-	modem->sms_manager->ops = ops;
-
-	if (!g_dbus_register_interface(conn, modem->path,
-					SMS_MANAGER_INTERFACE,
-					sms_manager_methods,
-					sms_manager_signals,
-					NULL, modem,
-					sms_manager_destroy)) {
-		ofono_error("Could not register SmsManager interface");
-		sms_manager_destroy(modem);
-
-		return -1;
-	}
-
-	ofono_debug("SmsManager interface for modem: %s created",
-			modem->path);
-
-	modem_add_interface(modem, SMS_MANAGER_INTERFACE);
+	g_drivers = g_slist_prepend(g_drivers, (void *)d);
 
 	return 0;
 }
 
-void ofono_sms_manager_unregister(struct ofono_modem *modem)
+void ofono_sms_driver_unregister(const struct ofono_sms_driver *d)
 {
-	DBusConnection *conn = dbus_gsm_connection();
+	DBG("driver: %p, name: %s", d, d->name);
 
-	g_dbus_unregister_interface(conn, modem->path,
-					SMS_MANAGER_INTERFACE);
+	g_drivers = g_slist_remove(g_drivers, (void *)d);
+}
 
-	modem_remove_interface(modem, SMS_MANAGER_INTERFACE);
+static void sms_unregister(struct ofono_atom *atom)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
+	const char *path = __ofono_atom_get_path(atom);
+
+	g_dbus_unregister_interface(conn, path, SMS_MANAGER_INTERFACE);
+	ofono_modem_remove_interface(modem, SMS_MANAGER_INTERFACE);
+}
+
+static void sms_remove(struct ofono_atom *atom)
+{
+	struct ofono_sms *sms = __ofono_atom_get_data(atom);
+
+	DBG("atom: %p", atom);
+
+	if (sms == NULL)
+		return;
+
+	if (sms->driver && sms->driver->remove)
+		sms->driver->remove(sms);
+
+	if (sms->tx_source) {
+		g_source_remove(sms->tx_source);
+		sms->tx_source = 0;
+	}
+
+	if (sms->assembly) {
+		sms_assembly_free(sms->assembly);
+		sms->assembly = NULL;
+	}
+
+	if (sms->txq) {
+		g_queue_foreach(sms->txq, (GFunc)g_free, NULL);
+		g_queue_free(sms->txq);
+		sms->txq = NULL;
+	}
+
+	g_free(sms);
+}
+
+struct ofono_sms *ofono_sms_create(struct ofono_modem *modem,
+					unsigned int vendor,
+					const char *driver,
+					void *data)
+{
+	struct ofono_sms *sms;
+	GSList *l;
+
+	if (driver == NULL)
+		return NULL;
+
+	sms = g_try_new0(struct ofono_sms, 1);
+
+	if (sms == NULL)
+		return NULL;
+
+	sms->sca.type = 129;
+	sms->ref = 1;
+	sms->txq = g_queue_new();
+	sms->atom = __ofono_modem_add_atom(modem, OFONO_ATOM_TYPE_SMS,
+						sms_remove, sms);
+
+	for (l = g_drivers; l; l = l->next) {
+		const struct ofono_sms_driver *drv = l->data;
+
+		if (g_strcmp0(drv->name, driver))
+			continue;
+
+		if (drv->probe(sms, vendor, data) < 0)
+			continue;
+
+		sms->driver = drv;
+		break;
+	}
+
+	return sms;
+}
+
+static void mw_watch(struct ofono_atom *atom,
+			enum ofono_atom_watch_condition cond, void *data)
+{
+	struct ofono_sms *sms = data;
+
+	if (cond == OFONO_ATOM_WATCH_CONDITION_UNREGISTERED) {
+		sms->mw = NULL;
+		return;
+	}
+
+	sms->mw = __ofono_atom_get_data(atom);
+}
+
+void ofono_sms_register(struct ofono_sms *sms)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_modem *modem = __ofono_atom_get_modem(sms->atom);
+	const char *path = __ofono_atom_get_path(sms->atom);
+	struct ofono_atom *mw_atom;
+	struct ofono_atom *sim_atom;
+
+	if (!g_dbus_register_interface(conn, path,
+					SMS_MANAGER_INTERFACE,
+					sms_manager_methods,
+					sms_manager_signals,
+					NULL, sms, NULL)) {
+		ofono_error("Could not create %s interface",
+				SMS_MANAGER_INTERFACE);
+		return;
+	}
+
+	ofono_modem_add_interface(modem, SMS_MANAGER_INTERFACE);
+
+	sms->mw_watch = __ofono_modem_add_atom_watch(modem,
+					OFONO_ATOM_TYPE_MESSAGE_WAITING,
+					mw_watch, sms, NULL);
+
+	mw_atom = __ofono_modem_find_atom(modem,
+					OFONO_ATOM_TYPE_MESSAGE_WAITING);
+
+	if (mw_atom && __ofono_atom_get_registered(mw_atom))
+		mw_watch(mw_atom, OFONO_ATOM_WATCH_CONDITION_REGISTERED, sms);
+
+	sim_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_SIM);
+
+	/* If we have a sim atom, we can uniquely identify the SIM,
+	 * otherwise create an sms assembly which doesn't backup the fragment
+	 * store.
+	 */
+	if (sim_atom) {
+		const char *imsi;
+
+		sms->sim = __ofono_atom_get_data(sim_atom);
+		imsi = ofono_sim_get_imsi(sms->sim);
+		sms->assembly = sms_assembly_new(imsi);
+	} else
+		sms->assembly = sms_assembly_new(NULL);
+
+	__ofono_atom_register(sms->atom, sms_unregister);
+}
+
+void ofono_sms_remove(struct ofono_sms *sms)
+{
+	__ofono_atom_free(sms->atom);
+}
+
+void ofono_sms_set_data(struct ofono_sms *sms, void *data)
+{
+	sms->driver_data = data;
+}
+
+void *ofono_sms_get_data(struct ofono_sms *sms)
+{
+	return sms->driver_data;
 }
